/* SPDX-License-Identifier: GPL-2.0 OR BSD-3-Clause */
/* Copyright(c) 2007 - 2011 Realtek Corporation. */

#ifndef _RTW_IOCTL_H_
#define _RTW_IOCTL_H_

#include "osdep_service.h"
#include "drv_types.h"

extern struct iw_handler_def  rtw_handlers_def;
extern int ui_pid[3];

#endif /*  #ifndef __INC_CEINFO_ */
