// SPDX-License-Identifier: ISC
/*
 * Copyright (c) 2012 Qualcomm Atheros, Inc.
 */

#include <linux/module.h>

#define CREATE_TRACE_POINTS
#include "trace.h"
EXPORT_SYMBOL(__tracepoint_ath10k_log_dbg);
