/* SPDX-License-Identifier: GPL-2.0 OR BSD-3-Clause */
/* Copyright(c) 2018-2019  Realtek Corporation
 */

#ifndef __RTW8723D_TABLE_H__
#define __RTW8723D_TABLE_H__

extern const struct rtw_table rtw8723d_mac_tbl;
extern const struct rtw_table rtw8723d_agc_tbl;
extern const struct rtw_table rtw8723d_bb_tbl;
extern const struct rtw_table rtw8723d_bb_pg_tbl;
extern const struct rtw_table rtw8723d_rf_a_tbl;
extern const struct rtw_table rtw8723d_txpwr_lmt_tbl;

#endif
